package PureMessage::Policy::Spam;

use strict;

use base 'PureMessage::Policy::Plugin';

use ActiveState::SpamEngine qw(PMX_TYPE_INTEGER
			       PMX_EV_FOUND_FEATURE PMX_EV_SPAMPROB);
use ActiveState::SpamEngine::Message;
use ActiveState::SpamEngine::Utils qw(unpack_summary);
use PureMessage::Logger;
use PureMessage::Policy::Spam::Engine qw(datadir new_engine new_asdb);
use PureMessage::Policy::Delay qw(add_sender_history_header add_rescan_header update_sender_history_counters);
use PureMessage::Policy::DelayMail qw(pmx_suspect_delay pmx_delayed_mail);
use PureMessage::Policy::Delay::Config qw(%config);
our $VERSION = '6.401';

# Shim PM::ListManager if we don't have it
{
    package PureMessage::Policy::Spam::ListManager;
    our @ISA;
    if (eval { require PureMessage::ListManager }) {
	push @ISA, qw(PureMessage::ListManager);
    }
    else {
	require PureMessage::Policy::Spam::ListManager_v4;
	push @ISA, qw(PureMessage::Policy::Spam::ListManager_v4);
    }
}

# Shim get_matcher() if we don't have PM::Policy::Plugins::Utils
if (eval { require PureMessage::Policy::Plugin::Utils }) {
    PureMessage::Policy::Plugin::Utils->import("get_matcher");
}
else {
    # This is how we used to do it in 4.x, where x < 5
    *get_matcher = sub {
	my ($msg, $args) = @_;
	return $args->matcher;
    };
}

# If PureMessage::Policy::Blocklist is available use it
{
    eval { require PureMessage::Policy::Blocklist };
}

# If PureMessage::Policy::DKIM is available use it
{
    eval { require PureMessage::Policy::DKIM };
}

# Disable pmx_offensive_word if we can't load PM::Policy::Words
my $have_offensive_word = eval { require PureMessage::Policy::Words };

if ($PureMessage::Milter::Policy::VERSION) {
    # running under milter: make sure all modules we might need
    # are loaded early so that the code can be shared between
    # all the milter children.
    require HTML::Parser;
    require Digest::MD5;

    # If we notice data staleness, compile the anti-spam data. This is
    # normally done by the various cronjobs that update the system, but if
    # they've used the Manager to modify the configuration, that won't be
    # picked up until now. Here's the algorithm:
    #
    # - `CHANGES'
    #   This file is updated by pmx-spam-compile.  The milter
    #   reloads if the timestamp on this file is different from
    #   the last time it was loaded.
    #
    # - `lock'
    #   This file is a hardlink to F<CHANGES> if the data is up-to-date.
    #   If not, the data needs to be recompiled.  We rely on the fact
    #   that the hardlink is invalidated by the Manager and the
    #   Publications system.
    my $ddir = datadir();
    my @chng = stat(my $chng = "$ddir/CHANGES");
    my @lock = stat(my $lock = "$ddir/lock");
    if (@lock and @chng and $lock[1] != $chng[1]) {
	unlink $lock; # remove stale lock if it exists
    }

    # If something fails, don't stop the milter.
    if (link($chng, $lock)) {
	my $cc = PureMessage::Config::expand_filename(
	    "bin/pmx-spam-compile"
	);
	system($cc) == 0 or warn "can't run $cc: \$?=$? \$!=$!\n";
    }
}

# Because of how Perl threading works, we can just use a global to store the
# current message being operated on. Each thread gets its own $msg. Until the
# first message actually shows up, we'll assign this a logger object instead.
our $msg_logger = PureMessage::Logger->new(ident => 'pmx_spam_prob');
my $eng;

sub logger {
    my ($level, $msg) = @_;
    $msg_logger->log($level, $msg);
}

my @pmx_version = qw(Antispam-Engine fix Antispam-Data fix);
sub commands {
    my ($cls, undef, $logger) = @_;
    my %cmds;

    use ActiveState::SpamEngine qw(PMX_TYPE_HASH);
    local $msg_logger = $logger if defined $logger;
    $eng = new_engine(logger => \&logger, compile => 0);
    my %p;
    $eng->enumerate_loaded_plugins(sub { $_[0]{$_[1]} = 1 }, \%p);
    if ($p{net}) {
	$eng->add_attribute(
	    "plugin.net.internal-hosts",
	    PMX_TYPE_HASH,
	    bless([], 'PureMessage::Policy::Spam::InternalHosts'),
	    1,
	);
	$eng->add_attribute(
	    "plugin.net.trusted-relays",
	    PMX_TYPE_HASH,
	    bless([], 'PureMessage::Policy::Spam::TrustedRelays'),
	    1,
	);
    }
    $eng->compile;

    %cmds = (
	pmx_spam_prob      => \&spamprob,
	pmx_spam_hit       => \&spamhit,
        pmx_suspect_delay  => \&PureMessage::Policy::DelayMail::pmx_suspect_delay,
        pmx_delayed_mail   => \&PureMessage::Policy::DelayMail::pmx_delayed_mail,

	# Check whether the offensive-words should be single or multi
	$have_offensive_word ? (pmx_offensive_word => \&offensive_word) : (),

	# If the policy engine supports it, provide version information
	$PureMessage::Policy::Engine::VERSION ge "4.5"
	    ? (__version__ => \@pmx_version)
	    : (),
    );

    # If the Blocklist package is installed allow all its commands to
    # be accessible via this package.
    eval {
        my %lookup = PureMessage::Policy::Blocklist->commands();
        foreach my $cmd (keys %lookup) {
            $cmds{$cmd} = $lookup{$cmd};
        }
    };
    # If the DKIM package is installed allow all its commands to
    # be accessible via this package.
    eval {
        my %lookup = PureMessage::Policy::DKIM->commands();
        foreach my $cmd (keys %lookup) {
            $cmds{$cmd} = $lookup{$cmd};
        }
    };
    %cmds;
}

# Grab the summary
my $summary = new_asdb(type => 'summary');

# Lists, and the refresh logic
my $lm;
{
    my $last_refresh = 0;
    my %match;
    sub REFRESH_INTERVAL() { 300 }
    sub refresh_lists {
	if (time() - $last_refresh > REFRESH_INTERVAL) {
	    $lm->refresh_lists;
	    $match{i} = $lm->find_list('internal-hosts')->matcher;
	    $match{t} = $lm->find_list('trusted-relays')->matcher;
	    $last_refresh = time();
	}
    }

    my @lists = qw(internal-hosts trusted-relays);
    $lm = PureMessage::Policy::Spam::ListManager->new(preload => \@lists);
    for my $list (@lists) {
	die "Error: '$list' is not allowed to be a multilist"
	    if $lm->is_multilist($list);
    }
    refresh_lists();

    sub pkg_version {
	my $f = PureMessage::Config::expand_filename("etc/.packages/$_[0]");
	open (my $FH, $f) or return;
	my $para = do { local $/ = ""; <$FH> };
	if ($para =~ /^Version: (.*)$/m) {
	    return $1;
	}
	undef;
    }

    $pmx_version[1] = pkg_version("PureMessage-AntiSpam-Engine");
    my $data_pkg = "PureMessage-AntiSpam-Data";
    $pmx_version[3] = pkg_version($data_pkg);

    my $last_recompile = time;
    my $changes = datadir() . "/CHANGES";
    sub check_recompile {
	my $mtime = (stat $changes)[9] or return;
	if ($mtime != $last_recompile) {
	    $pmx_version[3] = pkg_version($data_pkg);
	    $eng->log('INFO', "pmx_spam: loading new data version $pmx_version[3]");
	    $eng->add_attribute("core.compile", PMX_TYPE_INTEGER, 1, 1);
	    $last_recompile = $mtime;
	    $summary = new_asdb(type => 'summary');
        }
    }

# Pass in the internal-hosts list as "plugin.net.internal-hosts".

    package PureMessage::Policy::Spam::InternalHosts;
    sub get {
	my ($self, $host) = @_;
	return $match{i}->($host);
    }
    package PureMessage::Policy::Spam::TrustedRelays;
    sub get {
	my ($self, $host) = @_;
	return $match{t}->($host);
    }
}

sub extended_report {
    my ($hits, $x, $w) = @_;
    my @data = map {
		   ["$_$w->{$_}[2]", # name
		    $w->{$_}[0],     # weight
		    $w->{$_}[1],     # pdelta
		    $w->{$_}[3],     # description
		    $x->{$_}         # extra info (capture, uri...)
		   ]
	       }
	       grep { ! /^__/ }
	       @$hits;
    return "No antispam rules were triggered by this message" unless @data;

    # Maximum width
    sub MAX_FIELD_WIDTH() { 96 }

    # Find the longest rule name. Truncate at 30 characters.
    my $maxlen = 0;
    $maxlen < length($_->[0]) and $maxlen = length($_->[0]) for @data;
    $maxlen = 30 if $maxlen > 30;
    my $desclen = MAX_FIELD_WIDTH - 16 - $maxlen;

    my $fmt1 = sprintf("\t\@%s @#.### ^%s\n",
		       "<" x ($maxlen-1),
		       "<" x ($desclen-1));
    my $pad2 = MAX_FIELD_WIDTH - 8 - $desclen;
    my $fmt2 = "\t";
    while ($pad2 >= 8) { $fmt2 .= "\t", $pad2 -= 8 }
    while ($pad2 >  0) { $fmt2 .= " ", --$pad2 }
    $fmt2 .= "^" . "<" x ($desclen-1);
    $fmt2 .= "\n";

    my @extended;
    push @extended,
	"The following antispam rules were triggered by this message:\n";
    push @extended, sprintf("\t%-${maxlen}s  Score Description\n", 'Rule');
    foreach my $rule (@data) {
	local $^A = '';

	my $d = defined $rule->[3];
	my $x = defined $rule->[4];
	my $desc = '';
	if ($d && $x)   { $desc = "$rule->[3]: $rule->[4]" }
	elsif ($d)      { $desc = $rule->[3]               }
	elsif ($x)      { $desc = $rule->[4]               }

	# Escape non-printable characters.
	$desc =~ s/\r/\\r/g;
	$desc =~ s/\n/\\n/g;
	$desc =~ s/\t/\\t/g;
	$desc =~ s/([\x00-\x1f\x7f-\xff])/sprintf('\x%02x', ord($1))/eg;

	formline($fmt1, @$rule[0, 1], $desc);
	formline($fmt2, $desc) while length $desc;

	push @extended, $^A;
    }
    chomp $extended[-1];
    return join("", @extended);
}

sub probability {
    my ($msg) = @_;

    my $probability = $msg->get_attr('spamprob');
    my $gauge = $probability;
    unless (defined $probability) {
	my (@hits, %feature);
	# <delay_queue>
	if ((lc($config{delay_status}) ne 'off') && ($msg->minfo_value('Direction') eq 'Inbound')) {
	    if ($msg->get_attr('mark_dq_reinjected')) { # this will never be true if delay queue mode is 'Collect'
		add_rescan_header($msg);
	    }
	    else {
		# this will also add a new record for this 'Fur' if it is not already present in the historian DB
		add_sender_history_header($msg);
	    }
	}
	# </delay_queue>
	local $msg_logger = $msg;
	my $wrapper = ActiveState::SpamEngine::Message->new($msg);
	check_recompile();
	refresh_lists();
	$eng->scan_message($wrapper, undef, sub {
	    my (undef, $event, $arg1, $arg2) = @_;
	    if ($event == PMX_EV_SPAMPROB) {
		$probability = $arg1;
	    }
	    elsif ($event == PMX_EV_FOUND_FEATURE) {
		push @hits, $arg1;
		$feature{$arg1} = $arg2;
	    }
	});	

	# Store the weight and rule descriptions in %w
	my (%weights, $pdelta);
	$pdelta = 0;
	for my $hit (@hits) {
	    my ($weight, $prob, $feat, $desc) = split /\0/, $summary->get($hit);
	    $weights{$hit} = [$weight, $prob, $feat, $desc];
	    $pdelta += $prob;
	}
	# We want the HITS line sorted such that flagged rules appear first,
	# followed by the default rules. Negative rules always come before
	# positive ones, since they're rarer. Then we sort by absolute value
	# so that the most important rules are at the front.
	@hits = sort {
	    !!$weights{$b}[2] <=> !!$weights{$a}[2] or 
	    $weights{$b}[0]<0 <=> $weights{$a}[0]<0 or
	    abs($weights{$b}[0]) <=> abs($weights{$a}[0]) or
	    $a cmp $b
	} @hits;
	$msg->set_attr('mark_h', \@hits);
	$msg->minfo->append_items('Spam-Hit' => $_) for @hits;

    my @hit_lines;
    for my $hit (@hits) {
        push @hit_lines, "$hit$weights{$hit}[2] $weights{$hit}[0]";
    } 

    # DEF35415
    # We assume than each of hit must be less than 900 chars long.
    my $count = 0;
    my $line_length = 0;

    # According to RFC 5322, characters on a line cannot be
    # more than 998, but we decided to use 900.
    my $max_per_line = 900;

    # process first line
    # second line must start with space otherwise it will be treated
    # as another tag
    my $hits = "\n " . shift @hit_lines;
    $line_length = length($hits);
    # process rest of lines
    for my $hit_line (@hit_lines) {
        my $length = length($hit_line);
        if (($line_length + 3 + $length) > $max_per_line) {
            my $tmp = ",\n $hit_line";
            $line_length = length $tmp;
            $hits .= $tmp;
        } else {
            my $tmp = ", $hit_line";
            $line_length += length ($tmp);
            $hits .= $tmp;
        }
    }
    $msg->template('HITS', $hits);

	my $ext = extended_report(\@hits, \%feature, \%weights);
	$msg->template('SPAM_REPORT', $ext);

	# Add the pdelta for each feature
	$probability += $pdelta / 100;
	if ($probability > 1) { $probability = 1 }
	elsif ($probability < 0) { $probability = 0 }	
	$msg->log('DEBUG', "pmx_spam_prob: p=$probability");

	my $rounded = sprintf("%.3f", $probability);
	$msg->set_attr('mark_p', $rounded);
	$msg->minfo->append_items('Spam-Probability' => $rounded);

	# $int uses the integer rounded probability to be
	# more accurate and to allow probabilities of 100%.
	my $int = 100 * sprintf("%.2f", $probability);
	$msg->template('PROB', sprintf('%d%%', $int));
	$msg->set_attr(spamprob => $int);	
	$probability = $int;

        # set the gauge using the probability
        $gauge = $probability;
        $msg->template('XGAUGE', "X" x ($gauge / 10));
        $msg->template('SGAUGE', "*" x ($gauge / 10));
        $msg->template('IGAUGE', "I" x ($gauge % 10));
	# <delay_queue>
	if ((lc($config{delay_status}) ne 'off') && ($msg->minfo_value('Direction') eq 'Inbound') && (!$msg->get_attr('mark_dq_reinjected'))) {
	    update_sender_history_counters($msg, $probability);
	}
	# </delay_queue>
    }
    return ($gauge, $probability);
}

sub spamprob {
    my ($msg, $args, $first_time) = @_;
    $msg->assert_event('eom');
    my ($gauge, $probability) = probability($msg);
    my $n = $args->param(0);
    my $c = ($gauge - $n) / 10 + 1; # need at least one '#'
    $msg->template('GAUGE', '#' x $c);
    return $probability > $n if $args->tag('over');
    return $probability < $n if $args->tag('under');
    $probability == $n;
}

sub spamhit {
    my ($msg, $args, $first_time) = @_;
    my $hits;
    for (0 .. 1) {
	$hits = $msg->get_attr('mark_h');
	last if $hits;
	spamprob($msg, $args, $first_time); # run the spam tests if they haven't been run
    }
    return unless $hits and @$hits;
    my $matcher = get_matcher($msg, $args);
    return $matcher->(@$hits);
}

sub offensive_word {
    my ($msg, $args) = @_;
    my $offensive_m;
    if ($lm->is_multilist('offensive-words')) {
	$offensive_m = get_matcher($msg, $args);
    }
    else {
	$offensive_m = $lm->find_list('offensive-words')->matcher;
    }
    eval {
	require PureMessage::Policy::Words;
	PureMessage::Policy::Words->import(qw(message_words));
	message_words(sub {
	    my (undef, undef, $words) = @_;
	    my $str = join(' ', @$words);
            my @rv = $offensive_m->($str);
            if ($rv[0]) {
                $msg->log('DEBUG', "POLICY: Offensive word '$rv[1]' found");
                $msg->minfo->append_items('Matched-Word' => $rv[1]);
                die '--matched--';
            }	    
	}, $msg, overlap => 5, per_call => 1000)
    };
    if ($@) {
	return 1 if $@ =~ /--matched--/;
	die;
    }
    return 0;
}

1;
